from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from inspect import isasyncgenfunction, isgeneratorfunction
from types import TracebackType
from typing import AsyncContextManager, Callable, ContextManager

from graia.broadcast.interfaces.dispatcher import (
    DispatcherInterface as DispatcherInterface,
)

from ..entities.decorator import Decorator
from ..entities.dispatcher import BaseDispatcher
from ..entities.exectarget import ExecTarget
from ..entities.signatures import Force
from ..interfaces.decorator import DecoratorInterface
from ..interfaces.dispatcher import DispatcherInterface as DispatcherInterface


class Depend(Decorator):
    pre = True

    exec_target: ExecTarget
    cache: bool = False

    raw: Callable

    def __init__(self, callable: Callable, *, cache=False):
        self.cache = cache
        self.raw = callable

        if isgeneratorfunction(callable) or isgeneratorfunction(getattr(callable, "__call__", None)):
            callable = contextmanager(callable)
        elif isasyncgenfunction(callable) or isasyncgenfunction(getattr(callable, "__call__", None)):
            callable = asynccontextmanager(callable)

        self.exec_target = ExecTarget(callable)

    async def target(self, interface: DecoratorInterface):
        cache: dict = interface.local_storage["_depend_cached_results"]
        if self.raw in cache:
            return cache[self.raw]

        result_tier1 = await interface.dispatcher_interface.broadcast.Executor(
            target=self.exec_target,
            dispatchers=interface.dispatcher_interface.dispatchers,
            depth=interface.dispatcher_interface.depth + 1,
        )
        stack: AsyncExitStack = interface.local_storage["_depend_lifespan_manager"]

        if isinstance(result_tier1, ContextManager):
            result = stack.enter_context(result_tier1)
            cache[self.raw] = result
        elif isinstance(result_tier1, AsyncContextManager):
            result = await stack.enter_async_context(result_tier1)
            cache[self.raw] = result
        else:
            result = result_tier1
            if self.cache:
                cache[self.raw] = result

        return Force(result)


class DependDispatcher(BaseDispatcher):
    async def beforeExecution(self, interface: DispatcherInterface):
        interface.local_storage["_depend_lifespan_manager"] = AsyncExitStack()
        interface.local_storage["_depend_cached_results"] = {}

    async def catch(self, interface: DispatcherInterface):
        return

    async def afterExecution(
        self,
        interface: DispatcherInterface,
        exception: Exception | None,
        tb: TracebackType | None,
    ):
        stack: AsyncExitStack = interface.local_storage["_depend_lifespan_manager"]
        await stack.__aexit__(type(exception) if exception is not None else None, exception, tb)



import contextlib
from typing import TYPE_CHECKING, Optional

from ..entities.dispatcher import BaseDispatcher
from ..entities.event import Dispatchable

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface


class ExceptionThrown(Dispatchable):
    exception: Exception
    event: Optional[Dispatchable]

    def __init__(self, exception: Exception, event: Optional[Dispatchable]) -> None:
        self.exception = exception
        self.event = event

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: "DispatcherInterface[ExceptionThrown]"):
            event: ExceptionThrown = interface.event
            if event.event is not None:
                with contextlib.suppress(TypeError):
                    if interface.name == "event" or isinstance(interface.event.event, interface.annotation):
                        return interface.event.event
                    if interface.name == "exception" or isinstance(interface.event.exception, interface.annotation):
                        return interface.event.exception


class EventExceptionThrown(ExceptionThrown):
    event: Dispatchable

    def __init__(self, exception: Exception, event: Dispatchable) -> None:
        self.exception = exception
        self.event = event


ExceptionThrowed = EventExceptionThrown  # backward compatibility



from contextlib import AsyncExitStack
from types import TracebackType
from typing import TYPE_CHECKING, Optional, cast

from ..entities.decorator import Decorator
from ..entities.dispatcher import BaseDispatcher
from ..entities.signatures import Force
from ..utilles import Ctx, run_always_await

if TYPE_CHECKING:
    from ..interfaces.dispatcher import DispatcherInterface


ctx_dei_returnvalue = Ctx("ctx_dei_returnvalue")


class DecoratorInterface(BaseDispatcher):
    """Broadcast Control 内部机制 Decorator 的具体管理实现"""

    @property
    def dispatcher_interface(self) -> "DispatcherInterface":
        from .dispatcher import DispatcherInterface

        return DispatcherInterface.ctx.get()

    @property
    def name(self):
        return self.dispatcher_interface.name

    @property
    def annotation(self):
        return self.dispatcher_interface.annotation

    @property
    def event(self):
        return self.dispatcher_interface.event

    @property
    def return_value(self):
        return ctx_dei_returnvalue.get()

    @property
    def local_storage(self):
        return self.dispatcher_interface.local_storage

    async def catch(self, interface: "DispatcherInterface"):
        if isinstance(interface.default, Decorator):
            decorator: Decorator = interface.default
            with ctx_dei_returnvalue.use(
                await interface.lookup_param(interface.name, interface.annotation, None) if not decorator.pre else None
            ):
                return Force(await run_always_await(decorator.target, self))

    async def afterExecution(
        self, interface: "DispatcherInterface", exception: Optional[Exception], tb: Optional[TracebackType]
    ):
        stack = cast("AsyncExitStack | None", interface.local_storage.get("_depend_astack"))

        if stack is None:
            return

        if exception:
            await stack.__aexit__(type(exception), exception, tb)
        else:
            await stack.aclose()



import typing
from typing import Any, Optional, Union

from ..entities.decorator import Decorator
from ..entities.signatures import Force
from ..exceptions import RequirementCrashed
from ..interfaces.decorator import DecoratorInterface
from .depend import Depend as Depend


class OptionalParam(Decorator):
    pre = True

    def __init__(self, origin: Any):
        self.origin = origin

    async def target(self, interface: DecoratorInterface) -> Optional[Any]:
        annotation = interface.annotation
        if typing.get_origin(annotation) is Union:
            annotation = Union[tuple(x for x in typing.get_args(annotation) if x not in (None, type(None)))]  # type: ignore
        try:
            return Force(
                await interface.dispatcher_interface.lookup_by_directly(
                    interface,
                    interface.dispatcher_interface.name,
                    annotation,
                    self.origin,
                )
            )
        except RequirementCrashed:
            return Force()



from __future__ import annotations

import asyncio
from types import TracebackType
from typing import Awaitable, Callable

from ..entities.dispatcher import BaseDispatcher
from ..interfaces.dispatcher import DispatcherInterface


class DeferDispatcher(BaseDispatcher):
    async def beforeExecution(self, interface: DispatcherInterface):
        interface.local_storage["defer_callbacks"] = []

    async def catch(self, interface: DispatcherInterface):
        return

    async def afterExecution(
        self,
        interface: DispatcherInterface,
        exception: Exception | None,
        tb: TracebackType | None,
    ):
        callbacks: list[Callable[[DispatcherInterface, Exception | None, TracebackType | None], Awaitable[None]]]
        callbacks = interface.local_storage.get("defer_callbacks")  # type: ignore

        if not callbacks:
            return

        await asyncio.wait([i(interface, exception, tb) for i in callbacks])



from __future__ import annotations

from creart import AbstractCreator, CreateTargetInfo, it

from . import Broadcast
from .interrupt import InterruptControl


class BroadcastCreator(AbstractCreator):
    targets = (
        CreateTargetInfo(
            module="graia.broadcast",
            identify="Broadcast",
            humanized_name="Broadcast Control",
            description="<common,graia,broadcast> a high performance, highly customizable, elegantly designed event system based on asyncio",
            author=["GraiaProject@github"],
        ),
        CreateTargetInfo(
            module="graia.broadcast.interrupt",
            identify="InterruptControl",
            humanized_name="Interrupt",
            description="<common,graia,broadcast,interrupt> Interrupt feature for broadcast control.",
            author=["GraiaProject@github"],
        ),
    )

    @staticmethod
    def create(
        create_type: type[Broadcast | InterruptControl],
    ) -> Broadcast | InterruptControl:
        if issubclass(create_type, Broadcast):
            return create_type()
        elif issubclass(create_type, InterruptControl):
            return create_type(it(Broadcast))
